/*-----------------------------------------------------------------------------
/ Title      : iMS Test Program Compensation Table Sub-Menu Source
/ Project    : Isomet Modular Synthesiser System
/------------------------------------------------------------------------------
/ File       : $URL: http://nutmeg/svn/sw/trunk/09-Isomet/iMS_SDK/console_app/SubMenus/CompensationTableSubMenu.cpp $
/ Author     : $Author: dave $
/ Company    : Isomet (UK) Ltd
/ Created    : 2017-09-06
/ Last update: $Date: 2017-09-11 23:55:34 +0100 (Mon, 11 Sep 2017) $
/ Platform   :
/ Standard   : C++11
/ Revision   : $Rev: 300 $
/------------------------------------------------------------------------------
/ Description:
/------------------------------------------------------------------------------
/ Copyright (c) 2017 Isomet (UK) Ltd. All Rights Reserved.
/------------------------------------------------------------------------------
/ Revisions  :
/ Date        Version  Author  Description
/ 2017-09-06  1.0      dc      Created
/
/----------------------------------------------------------------------------*/

#include "CompensationTableSubMenu.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <atomic>
#include <cmath>
#include <chrono>
#include <thread>

#include "IEventHandler.h"
#include "IMSTypeDefs.h"

static std::string desc("Program a Compensation Table");

class CompensationTableDownloadSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_downloading{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::CompensationEvents::DOWNLOAD_FINISHED) : std::cout << "Comp Table Download Finished!" << std::endl; m_downloading.store(false); break;
		case (iMS::CompensationEvents::DOWNLOAD_ERROR) : std::cout << "Download error on message " << param << std::endl; break;
		}
	}
	bool Busy() const { return m_downloading.load(); };
};

class CompensationTableVerifySupervisor : public iMS::IEventHandler
{
private:
	bool m_verify_success{ false };
	std::atomic<bool> m_verifying{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::CompensationEvents::VERIFY_SUCCESS) : std::cout << "Comp Table Verify Successful!" << std::endl; m_verify_success = true; m_verifying.store(false); break;
		case (iMS::CompensationEvents::VERIFY_FAIL) : std::cout << "Comp Table Verify FAILED!" << std::endl; m_verifying.store(false); break;
		}
	}
	bool Busy() const { return m_verifying.load(); };
	bool Success() const { return m_verify_success; };
};

bool WaitForCompensationDownload(iMS::IMSSystem& ims, const iMS::CompensationTable& tbl)
{
	iMS::CompensationTableDownload * dl = new iMS::CompensationTableDownload(ims, tbl);
	CompensationTableDownloadSupervisor ds;

	dl->CompensationTableDownloadEventSubscribe(iMS::CompensationEvents::DOWNLOAD_FINISHED, &ds);
	dl->CompensationTableDownloadEventSubscribe(iMS::CompensationEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		std::cout << "Error starting download" << std::endl;
		dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::DOWNLOAD_FINISHED, &ds);
		dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::DOWNLOAD_ERROR, &ds);
		delete dl;
		return false;
	}

	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}

	dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::DOWNLOAD_FINISHED, &ds);
	dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::DOWNLOAD_ERROR, &ds);

	CompensationTableVerifySupervisor vs;
	dl->CompensationTableDownloadEventSubscribe(iMS::CompensationEvents::VERIFY_SUCCESS, &vs);
	dl->CompensationTableDownloadEventSubscribe(iMS::CompensationEvents::VERIFY_FAIL, &vs);

	if (!dl->StartVerify()) {
		std::cout << "Error starting verify" << std::endl;
		dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::VERIFY_SUCCESS, &vs);
		dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::VERIFY_FAIL, &vs);
		delete dl;
		return false;
	}

	while (vs.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}
	if (!vs.Success()) {
		std::cout << "CT Unprogrammed" << std::endl;
	}
	dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::VERIFY_SUCCESS, &vs);
	dl->CompensationTableDownloadEventUnsubscribe(iMS::CompensationEvents::VERIFY_FAIL, &vs);

	int error;
	while ((error = dl->GetVerifyError()) != -1)
	{
		std::cout << "Failure around compensation table point: " << error << std::endl;
	}

	delete dl;
	return true;
}

const int CompensationTableSubMenu::num_submenu_items = 4;

CompensationTableSubMenu::CompensationTableSubMenu(iMS::IMSSystem& ims) :
				thisIMS(ims),
				comp(iMS::CompensationTable(ims))
{

}

CompensationTableSubMenu::~CompensationTableSubMenu() {
}

const std::string& CompensationTableSubMenu::SubMenuDescription() const
{
	return desc;
}

void CompensationTableSubMenu::SubMenuPrint()
{
	std::cout << std::endl << std::endl;
	std::cout << " ******** COMPENSATION TABLE MENU ********" << std::endl;
	std::cout << " 1) Create an Example Compensation Table" << std::endl;
	std::cout << " 2) Load CT from .LUT file" << std::endl;
	std::cout << " 3) Download to iMS System (volatile)" << std::endl;
	std::cout << " 4) Store on iMS FileSystem (non-volatile)" << std::endl;
	std::cout << " q) Back to Main Menu" << std::endl;
}

void CompensationTableSubMenu::LaunchSubMenu()
{
	std::string s;

	do {
		this->SubMenuPrint();
		std::getline( std::cin, s);
		if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
			break;
		} else {
			int index;
			std::istringstream iss(s);
			iss >> index;
			if (!iss.fail()) {
				if ((index > 0) && (index <= num_submenu_items)) {
					switch(index) {
					case 1: {
						// Create an Example CT
						std::string s;
						std::cout << "Generate an example Compensation Table to program the iMS LUT." << std::endl;
						std::cout << "Example creates an amplitude curve starting at Frequency F1, ending at" << std::endl;
						std::cout << "F2 with Amplitude A.\n\n" ;
						std::cout << "The phase will sweep from 0 to 360 between F1 and F2 as will the analog " << std::endl;
						std::cout << "and digital outputs.\n\n" ;
						std::cout << "The result is artificial in that it has little practical value but it does" << std::endl;
						std::cout << "demonstrate the process for manually creating your own Compensation Tables\n\n";

						iMS::MHz F1(0.0), F2(0.0);
						iMS::Percent A;
						double d;

						do {
							std::cout << "Enter Start Frequency F1 (MHz): ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> d;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else {
								F1 = d;
								break;
							}
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						do {
							std::cout << "Enter Start Frequency F2 (MHz): ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> d;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else {
								F2 = d;
								break;
							}
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						do {
							std::cout << "Enter Final Amplitude A (%): ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> d;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else {
								A = d;
								break;
							}
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						int i = 0;
						for (auto&& elem : comp)
						{
							d = comp.FrequencyAt(i);

							if (d < F1) {
								elem.Amplitude(iMS::Percent(100.0));
								elem.Phase(iMS::Degrees(0.0));
								elem.SyncAnlg(1.0f);
							} else if (d > F2) {
								elem.Amplitude(A);
								elem.Phase(iMS::Degrees(360.0));
								elem.SyncAnlg((double)A / 100.0f);
							} else {
								elem.Phase(360.0 * (d/(double)(F2 - F1)));
								d = (1.0/(std::pow(10,(2.0*(d - F1)/(double)(F2 - F1)))));
								elem.Amplitude(d*(100.0 - A) + A);
								elem.SyncAnlg(d);
							}
							elem.SyncDig(i);

							i++;
						}

						break;
					}
					case 2: {
						// Load CT from .LUT file
						std::string filename;
						std::cout << "Name of local file to send: ";
						std::getline( std::cin, filename);

						// open the file:
						std::ifstream file(filename.c_str(), std::ios::binary);

						if (file.good()) {
							comp = iMS::CompensationTable(thisIMS, filename);
						} else {
							std::cout << "Can't open " << filename << std::endl;
						}

						break;
					}
					case 3: {
						// Download to iMS
						std::cout << "Programming Compensation Table into iMS . . . (please wait)" << std::endl;
						WaitForCompensationDownload(thisIMS, comp);
						break;
					}
					case 4: {
						// Store to FileSystem
						std::string filename, def;
						std::cout << "Please enter a Filesystem associated name (max 8 chars): " ;
						std::getline( std::cin, filename);
						std::cout << "\nIs default? (programs LUT at startup) [y|N]: " << std::endl;
						std::getline( std::cin, def);

						iMS::CompensationTableDownload ctdl(thisIMS, comp);
						if ((def.substr(0,1).find("y") != std::string::npos) || (def.substr(0,1).find("Y") != std::string::npos)) {
							ctdl.Store(iMS::FileDefault::DEFAULT, filename);
						} else {
							ctdl.Store(iMS::FileDefault::NON_DEFAULT, filename);
						}

						break;
					}
					}
				}
			}
		}
	} while(1);

}

