/*-----------------------------------------------------------------------------
/ Title      : iMS Test Program Image Operations Sub-Menu Source
/ Project    : Isomet Modular Synthesiser System
/------------------------------------------------------------------------------
/ File       : $URL: http://nutmeg/svn/sw/trunk/09-Isomet/iMS_SDK/console_app/SubMenus/ImageSubMenu.cpp $
/ Author     : $Author: dave $
/ Company    : Isomet (UK) Ltd
/ Created    : 2017-09-06
/ Last update: $Date: 2017-09-11 23:55:34 +0100 (Mon, 11 Sep 2017) $
/ Platform   :
/ Standard   : C++11
/ Revision   : $Rev: 300 $
/------------------------------------------------------------------------------
/ Description:
/------------------------------------------------------------------------------
/ Copyright (c) 2017 Isomet (UK) Ltd. All Rights Reserved.
/------------------------------------------------------------------------------
/ Revisions  :
/ Date        Version  Author  Description
/ 2017-09-06  1.0      dc      Created
/
/----------------------------------------------------------------------------*/

#include "ImageSubMenu.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <atomic>
#include <chrono>
#include <thread>

#include "IMSTypeDefs.h"
#include "ImageProject.h"

static std::string desc("Image Mode Operations");

class ImageDownloadSupervisor : public iMS::IEventHandler
{
private:
	bool m_downloading{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImageDownloadEvents::DOWNLOAD_FINISHED) : std::cout << "Image Download Finished!" << std::endl; m_downloading = false; break;
		case (iMS::ImageDownloadEvents::DOWNLOAD_ERROR) :/* std::cout << "Download error on message " << param << std::endl;*/ break;
		}
	}
	bool Busy() const { return m_downloading; };
};

class ImageVerifySupervisor : public iMS::IEventHandler
{
private:
	bool m_verifying{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImageDownloadEvents::VERIFY_SUCCESS) : std::cout << "Image Verify Successful!" << std::endl; m_verifying = false; break;
		case (iMS::ImageDownloadEvents::VERIFY_FAIL) : std::cout << "Image Verify FAILED! First error byte: " << param << std::endl; m_verifying = false; break;
		}
	}
	bool Busy() const { return m_verifying; };
};

class PlaybackSupervisor : public iMS::IEventHandler
{
private:
	bool m_playing{ false };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImagePlayerEvents::IMAGE_STARTED) : std::cout << "Image Playback Started" << std::endl; m_playing = true; break;
		case (iMS::ImagePlayerEvents::POINT_PROGRESS) : std::cout << "Progress: " << param << std::endl; break;
		case (iMS::ImagePlayerEvents::IMAGE_FINISHED) : std::cout << "Image Playback Finished" << std::endl; m_playing = false; break;
		}
	}
	bool Busy() const { return m_playing; };
};

bool WaitForImageDownload(iMS::IMSSystem& ims, const iMS::Image& img)
{
	iMS::ImageDownload * dl = new iMS::ImageDownload(ims, img);
	ImageDownloadSupervisor ds;

	dl->ImageDownloadEventSubscribe(iMS::ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventSubscribe(iMS::ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		std::cout << "Error starting download" << std::endl;
		dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
		dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::DOWNLOAD_ERROR, &ds);
		delete dl;
		return false;
	}

	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}

	dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	ImageVerifySupervisor vs;
	dl->ImageDownloadEventSubscribe(iMS::ImageDownloadEvents::VERIFY_SUCCESS, &vs);
	dl->ImageDownloadEventSubscribe(iMS::ImageDownloadEvents::VERIFY_FAIL, &vs);

	if (!dl->StartVerify()) {
		std::cout << "Error starting verify" << std::endl;
		dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::VERIFY_SUCCESS, &vs);
		dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::VERIFY_FAIL, &vs);
		delete dl;
		return false;
	}

	while (vs.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}

	dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::VERIFY_SUCCESS, &vs);
	dl->ImageDownloadEventUnsubscribe(iMS::ImageDownloadEvents::VERIFY_FAIL, &vs);

	int error;
	while ((error = dl->GetVerifyError()) != -1)
	{
		std::cout << "Failure around image point: " << error << std::endl;
	}

	delete dl;
	return true;
}

const int ImageSubMenu::num_submenu_items = 6;

ImageSubMenu::ImageSubMenu(iMS::IMSSystem& ims) :
			thisIMS(ims),
			img_count(0),
			itv(new iMS::ImageTableViewer(ims))
{

}

ImageSubMenu::~ImageSubMenu() {
	delete itv;
}

const std::string& ImageSubMenu::SubMenuDescription() const
{
	return desc;
}

void ImageSubMenu::SubMenuPrint()
{
	std::cout << std::endl << std::endl;
	std::cout << " ******** IMAGE OPERATIONS MENU ********" << std::endl;
	std::cout << " 1) Create an Example Frequency Ramp Image" << std::endl;
	std::cout << " 2) Create an Example Sawtooth Image" << std::endl;
	std::cout << " 3) Load Image from Image Project" << std::endl;
	std::cout << " 4) Download Image" << std::endl;
	std::cout << " 5) Play Image" << std::endl;
	std::cout << " 6) Stop Image Playback" << std::endl;
	std::cout << " q) Back to Main Menu" << std::endl;
}

void ImageSubMenu::LaunchSubMenu()
{
	std::string s;

	do {
		this->SubMenuPrint();
		std::getline( std::cin, s);
		if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
			break;
		} else {
			int index;
			std::istringstream iss(s);
			iss >> index;
			if (!iss.fail()) {
				if ((index > 0) && (index <= num_submenu_items)) {
					switch(index) {
					case 1: {
						// Create an Example Frequency Ramp Image
						iMS::MHz F1(0.0), F2(0.0);
						double d;
						int n_pts;
						bool alternate = false;

						do {
							std::cout << "Enter Start Frequency F1 (MHz): ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> d;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else {
								F1 = d;
								break;
							}
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						do {
							std::cout << "Enter Start Frequency F2 (MHz): ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> d;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else {
								F2 = d;
								break;
							}
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						do {
							std::cout << "Enter Number of Image Points: ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> n_pts;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else break;
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						{
							std::cout << "Alternate channels? [y|N] ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.substr(0,1).find("y") != std::string::npos) || (s.substr(0,1).find("Y") != std::string::npos)) {
								alternate = true;
							}
						}

						std::stringstream img_name;
						img_name << "Ramp Example " << img_count++;
						img = iMS::Image(n_pts, iMS::ImagePoint(), img_name.str());

						iMS::FAP fap1(iMS::MHz(100.0), iMS::Percent(100.0), iMS::Degrees(0.0));
						iMS::FAP fap2(iMS::MHz(100.0), iMS::Percent(100.0), iMS::Degrees(0.0));
						int index = 0;
						for (auto&& pt : img)
						{
							fap1.freq = ((F2 - F1) * ((double)index / (n_pts - 1))) + F1;
							if (alternate) {
								fap2.freq = F2 - fap1.freq + F1;
							} else {
								fap2.freq = fap1.freq;
							}
							pt.SetFAP(iMS::RFChannel(1)) = fap1;
							pt.SetFAP(iMS::RFChannel(2)) = fap2;
							pt.SetFAP(iMS::RFChannel(3)) = fap1;
							pt.SetFAP(iMS::RFChannel(4)) = fap2;
							pt.SetSyncA(0, ((double)index++ / (n_pts - 1)));
						}
						break;
					}
					case 2: {
						// Create an Example Sawtooth Image
						iMS::MHz F1(0.0), F2(0.0);
						double d;
						int n_pts;
						bool alternate = false;

						do {
							std::cout << "Enter Start Frequency F1 (MHz): ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> d;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else {
								F1 = d;
								break;
							}
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						do {
							std::cout << "Enter Start Frequency F2 (MHz): ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> d;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else {
								F2 = d;
								break;
							}
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						do {
							std::cout << "Enter Number of Image Points: ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> n_pts;
							if (iss.fail()) {
								std::cout << "Try again - ";
								continue;
							} else break;
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						{
							std::cout << "Alternate channels? [y|N] ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.substr(0,1).find("y") != std::string::npos) || (s.substr(0,1).find("Y") != std::string::npos)) {
								alternate = true;
							}
						}

						std::stringstream img_name;
						img_name << "Sawtooth Example " << img_count++;
						img = iMS::Image(n_pts, iMS::ImagePoint(), img_name.str());

						iMS::FAP fap1(iMS::MHz(100.0), iMS::Percent(100.0), iMS::Degrees(0.0));
						iMS::FAP fap2(iMS::MHz(100.0), iMS::Percent(100.0), iMS::Degrees(0.0));
						int index = 0;
						for (auto&& pt : img)
						{
							fap1.freq = ((F2 - F1) * 2 * ((double)index / (n_pts - 1))) + F1;
							if (fap1.freq > F2) fap1.freq = 2*F2 - fap1.freq;
							if (alternate) {
								fap2.freq = F2 - fap1.freq + F1;
							} else {
								fap2.freq = fap1.freq;
							}
							pt.SetFAP(iMS::RFChannel(1)) = fap1;
							pt.SetFAP(iMS::RFChannel(2)) = fap2;
							pt.SetFAP(iMS::RFChannel(3)) = fap1;
							pt.SetFAP(iMS::RFChannel(4)) = fap2;
							pt.SetSyncA(0, ((double)index++ / (n_pts - 1)));
						}
						break;
					}
					case 3: {
						// Load Image from Image Project
						iMS::ImageProject proj;
						std::string filename;

						std::cout << "Enter filename of Image Project (*.xml): ";
						std::getline( std::cin, filename);

						// open the file:
						std::ifstream file(filename.c_str(), std::ios::binary);

						if (file.good()) {
							proj.Load(filename);
						} else {
							std::cout << "Can't open " << filename << std::endl;
						}

						std::cout << "\nImage Groups:\n";
						iMS::ImageGroupList& imglist = proj.ImageGroupContainer();
						unsigned int i=1;
						for (iMS::ImageGroupList::const_iterator it = imglist.cbegin(); it != imglist.cend(); ++it) {
							std::cout << "\n   " << it->Name() << ":" << std::endl;
							for (unsigned int j=1; j<= it->size(); j++) {
								std::cout << "\n    [" << i++ << "]: " << (*it)[j-1].Name();
							}
							std::cout << std::endl;
						}

						unsigned int img_idx;
						do {
							std::cout << "Select an Image : ";
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> img_idx;
							if (iss.fail() || (img_idx < 1) || (img_idx >= i)) {
								std::cout << "Try again - ";
								continue;
							} else break;
						} while(1);
						if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
							break;
						}

						i = 1;
						for (iMS::ImageGroupList::const_iterator it = imglist.cbegin(); it != imglist.cend(); ++it) {
							int grp_idx=0;
							for (unsigned int j=1; j<= it->size(); j++) {
								if (i++ == img_idx) {
									img = (*it)[grp_idx];
								}
								grp_idx++;
							}
						}
						break;
					}
					case 4: {
						// Download Image
						WaitForImageDownload(thisIMS, img);

						break;
					}
					case 5: {
						// Play Image
						int img_idx;
						double d;

						do {
							std::cout << "Select an Image from the Image Table to play ..." << std::endl;
							std::cout << *itv;
							std::getline( std::cin, s);
							iss.clear();
							iss.str(s);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}
							iss >> img_idx;
							if (iss.fail() || (img_idx < 0) || (img_idx >= itv->Entries())) {
								std::cout << "Try again - ";
								continue;
							} else break;
						} while(1);

						bool internal = true;
						std::cout << "Internal (default) or External Clock/Trigger? [I|e]: ";
						std::getline( std::cin, s);
						if ((s.find("e") != std::string::npos) || (s.find("E") != std::string::npos)) {
							internal = false;
						}

						iMS::kHz ClockRate(0.0);
						if (internal) {
							do {
								std::cout << "Enter Image Clock Rate (kHz): ";
								std::getline( std::cin, s);
								iss.clear();
								iss.str(s);
								if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
									break;
								}
								iss >> d;
								if (iss.fail()) {
									std::cout << "Try again - ";
									continue;
								} else {
									ClockRate = d;
									break;
								}
							} while(1);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}

							int delay;
							do {
								std::cout << "Enter Post-Image Delay (msec): ";
								std::getline( std::cin, s);
								iss.clear();
								iss.str(s);
								if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
									break;
								}
								iss >> delay;
								if (iss.fail()) {
									std::cout << "Try again - ";
									continue;
								} else {
									break;
								}
							} while(1);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}

							std::unique_ptr<iMS::ImagePlayer> player (new iMS::ImagePlayer(thisIMS, (*itv)[img_idx], ClockRate));
							player->cfg.rpts = iMS::ImagePlayer::Repeats::FOREVER;
							player->SetPostDelay(std::chrono::milliseconds(delay));
							player->cfg.trig = iMS::ImagePlayer::ImageTrigger::POST_DELAY;
							player->cfg.int_ext = iMS::ImagePlayer::PointClock::INTERNAL;

							if (!player->Play()) std::cout << "Playback failed" << std::endl;

						} else {
							int div;
							do {
								std::cout << "Enter External Clock Divider Ratio: ";
								std::getline( std::cin, s);
								iss.clear();
								iss.str(s);
								if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
									break;
								}
								iss >> div;
								if (iss.fail()) {
									std::cout << "Try again - ";
									continue;
								} else {
									break;
								}
							} while(1);
							if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
								break;
							}

							std::unique_ptr<iMS::ImagePlayer> player (new iMS::ImagePlayer(thisIMS, (*itv)[img_idx], div));
							player->cfg.rpts = iMS::ImagePlayer::Repeats::FOREVER;
							player->cfg.trig = iMS::ImagePlayer::ImageTrigger::EXTERNAL;
							player->cfg.int_ext = iMS::ImagePlayer::PointClock::EXTERNAL;

							if (!player->Play()) std::cout << "Playback failed" << std::endl;

						}

						break;
					}
					case 6: {
						// Stop Image Playback
						iMS::ImagePlayer ForceStop(thisIMS, img);  // Object instantiation requires 'img' but will stop any image that is playing
						ForceStop.Stop(iMS::ImagePlayer::StopStyle::IMMEDIATELY);

					}
					}
				}
			}
		}
	} while(1);

}


