/*-----------------------------------------------------------------------------
/ Title      : iMS Test Program Tone Buffer Sub-Menu Source
/ Project    : Isomet Modular Synthesiser System
/------------------------------------------------------------------------------
/ File       : $URL: http://nutmeg/svn/sw/trunk/09-Isomet/iMS_SDK/console_app/SubMenus/ToneBufferSubMenu.cpp $
/ Author     : $Author: dave $
/ Company    : Isomet (UK) Ltd
/ Created    : 2017-09-06
/ Last update: $Date: 2017-09-11 23:55:34 +0100 (Mon, 11 Sep 2017) $
/ Platform   :
/ Standard   : C++11
/ Revision   : $Rev: 300 $
/------------------------------------------------------------------------------
/ Description:
/------------------------------------------------------------------------------
/ Copyright (c) 2017 Isomet (UK) Ltd. All Rights Reserved.
/------------------------------------------------------------------------------
/ Revisions  :
/ Date        Version  Author  Description
/ 2017-09-06  1.0      dc      Created
/
/----------------------------------------------------------------------------*/

#include "ToneBufferSubMenu.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <atomic>
#include <chrono>
#include <thread>

#include "IMSTypeDefs.h"

static std::string desc("Tone Buffer Operations");

class ToneBufferDownloadSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_downloading{ true };
	std::atomic<bool> m_error{ false };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ToneBufferEvents::DOWNLOAD_FINISHED) : std::cout << "Tone Buffer Download Finished!" << std::endl; m_downloading.store(false); break;
		case (iMS::ToneBufferEvents::DOWNLOAD_ERROR) : m_error.store(true); break;
		}
	}
	bool Busy() const { return m_downloading.load(); };
	bool Error() const { return m_error.load(); };
	void Reset() { m_downloading.store(true); m_error.store(false); };
};

bool WaitForToneBufferDownload(iMS::IMSSystem& ims, const iMS::ToneBuffer& tb)
{
	iMS::ToneBufferDownload * dl = new iMS::ToneBufferDownload(ims, tb);
	ToneBufferDownloadSupervisor ds;

	dl->ToneBufferDownloadEventSubscribe(iMS::ToneBufferEvents::DOWNLOAD_FINISHED, &ds);
	dl->ToneBufferDownloadEventSubscribe(iMS::ToneBufferEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		std::cout << "Error starting download" << std::endl;
		dl->ToneBufferDownloadEventUnsubscribe(iMS::ToneBufferEvents::DOWNLOAD_FINISHED, &ds);
		dl->ToneBufferDownloadEventUnsubscribe(iMS::ToneBufferEvents::DOWNLOAD_ERROR, &ds);
		delete dl;
		return false;
	}

	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}

	dl->ToneBufferDownloadEventUnsubscribe(iMS::ToneBufferEvents::DOWNLOAD_FINISHED, &ds);
	dl->ToneBufferDownloadEventUnsubscribe(iMS::ToneBufferEvents::DOWNLOAD_ERROR, &ds);

	delete dl;
	if (ds.Error())
	{
		std::cout << "Tone Buffer Download FAILED. " << std::endl;
		return false;
	}

	return true;
}

const int ToneBufferSubMenu::num_submenu_items = 6;

ToneBufferSubMenu::ToneBufferSubMenu(iMS::IMSSystem& ims) :
	thisIMS(ims),
	tonebuf(new iMS::ToneBuffer()),
	sp(new iMS::SignalPath(ims))
{

}

ToneBufferSubMenu::~ToneBufferSubMenu() {
	delete sp;
	delete tonebuf;
}

const std::string& ToneBufferSubMenu::SubMenuDescription() const
{
	return desc;
}

void ToneBufferSubMenu::SubMenuPrint()
{
	std::cout << std::endl << std::endl;
	std::cout << " ******** TONE BUFFER MENU ********" << std::endl;
	std::cout << " 1) Create a Tone Buffer" << std::endl;
	std::cout << " 2) Download Tone Buffer" << std::endl;
	std::cout << " 3) Select a Tone" << std::endl;
	std::cout << " 4) Enable LUT Compensation" << std::endl;
	std::cout << " 5) Disable LUT Compensation" << std::endl;
	std::cout << " 6) Select Control Source" << std::endl;
	std::cout << " q) Back to Main Menu" << std::endl;
}

void ToneBufferSubMenu::LaunchSubMenu()
{
	std::string s;

	do {
		this->SubMenuPrint();
		std::getline( std::cin, s);
		if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
			break;
		} else {
			int index;
			std::istringstream iss(s);
			iss >> index;
			if (!iss.fail()) {
				if ((index > 0) && (index <= num_submenu_items)) {
					switch(index) {
					case 1: {
						// Create a Tone Buffer
						std::cout << "Generate an example Tone Buffer with linearly spaced tones between" << std::endl;
						std::cout << "the upper and lower frequency bounds of the synthesiser\n\n";
						std::cout << "Choose how many tone entries (up to 256) to create: " << std::endl;

						std::getline( std::cin, s);
						iss.clear();
						iss.str(s);
						int n_tones;
						iss >> n_tones;
						if ((n_tones >= 1) && (n_tones <= 256) && !(iss.fail())) {
							for (int i = 0; i < n_tones; i++)
							{
								(*tonebuf)[i].SetAll(iMS::FAP(iMS::MHz(thisIMS.Synth().GetCap().lowerFrequency + (((double)i / (n_tones - 1)) *
										(thisIMS.Synth().GetCap().upperFrequency - thisIMS.Synth().GetCap().lowerFrequency))), iMS::Percent(100.0), iMS::Degrees(0.0)));
							}
							for (int i = n_tones; i < 256; i++)
							{
								(*tonebuf)[i].SetAll(iMS::FAP(iMS::MHz(0.0), iMS::Percent(0.0), iMS::Degrees(0.0)));
							}
						} else {
							std::cout << "Please use an integer between 1 and 256" << std::endl;
						}
						break;
					}
					case 2: {
						// Download Tone Buffer
						if (!WaitForToneBufferDownload(thisIMS, *tonebuf)) {
							std::cout << "Tone Buffer Download Failed." << std::endl;
						} else {
							std::cout << "Tone Buffer Download Succeeded." << std::endl;
						}
						break;
					}
					case 3: {
						// Select a Tone
						std::cout << "Enter a tone index (1 - 256, or -1 to exit tone buffer mode): ";
						std::getline( std::cin, s);
						iss.clear();
						iss.str(s);
						int index;
						iss >> index;
						if (!iss.fail()) {
							if (-1 == index) {
								sp->UpdateLocalToneBuffer(iMS::SignalPath::ToneBufferControl::OFF);
							} else {
								sp->UpdateLocalToneBuffer(index);
							}
						}
						break;
					}
					case 4: {
						// Enable Compensation
						std::cout << "Tone Buffer LUT Amplitude & Phase Compensation ENABLED" << std::endl;
						sp->UpdateLocalToneBuffer(iMS::SignalPath::Compensation::ACTIVE,
								iMS::SignalPath::Compensation::ACTIVE);
						break;
					}
					case 5: {
						// Disable Compensation
						std::cout << "Tone Buffer LUT Amplitude & Phase Compensation DISABLED" << std::endl;
						sp->UpdateLocalToneBuffer(iMS::SignalPath::Compensation::BYPASS,
								iMS::SignalPath::Compensation::BYPASS);
						break;
					}
					case 6: {
						// Select Control Source
						std::cout << "Select Source: (1) = Software (Host), (2) = External input signal (up to 16 entries)" << std::endl <<
								"               (3) = Extended External input signal (up to 256 entries)" << std::endl;
						std::getline( std::cin, s);
						iss.clear();
						iss.str(s);
						iss >> index;
						if ((index >= 1) && (index <= 3)) {
							switch(index) {
							case 1: {
								sp->UpdateLocalToneBuffer(iMS::SignalPath::ToneBufferControl::HOST);
								std::cout << "Using Software Host Control" << std::endl;
								break;
							}
							case 2: {
								sp->UpdateLocalToneBuffer(iMS::SignalPath::ToneBufferControl::EXTERNAL);
								std::cout << "Using External Signal Control" << std::endl;
								break;
							}
							case 3: {
								sp->UpdateLocalToneBuffer(iMS::SignalPath::ToneBufferControl::EXTERNAL_EXTENDED);
								std::cout << "Using Extended External Signal Control" << std::endl;
								break;
							}
							}
						} else {
							std::cout << "Control source unchanged" << std::endl;
						}
						break;
					}

					}
				}
			}
		}
	} while(1);

}


