#include <cstdlib>
#include <iostream>
#include <sstream>
#include <list>
#include <memory>

#include "SubMenus/SubMenuList.h"

#include "ConnectionList.h"
#include "LibVersion.h"

using namespace iMS;

static const char WelcomeMsg[] = "Welcome to Isomet iMS on QNX.";

#if defined(__QNXNTO__)
#define CONSOLE_HEIGHT 25
#else
#define CONSOLE_HEIGHT 80
#endif

static SubMenuList* menus;

// A woefully pathetic way to clear the terminal output, but it works and is cross platform
// There aren't many "better" ways! (ncurses is archaic; system("clear") is BAD).
void ClearScreen()
{
	std::cout << std::string ( CONSOLE_HEIGHT, '\n');
}

void PrintIMSDetails(const IMSSystem& myiMS)
{
	if (myiMS.Open()) {
		std::cout << "Connected to IMS System on port: " << myiMS.ConnPort() << std::endl;;
		if (myiMS.Synth().IsValid()) {
			std::cout << "Found Synthesiser: " << myiMS.Synth().Model() << " : " <<  myiMS.Synth().Description() << std::endl;
			std::cout << " FW Version: " << myiMS.Synth().GetVersion() << std::endl;
		}
		if (myiMS.Ctlr().IsValid()) {
			std::cout << "Found Controller: " << myiMS.Ctlr().Model() << " : " << myiMS.Ctlr().Description() << std::endl;
			std::cout << " FW Version: " << myiMS.Ctlr().GetVersion() << std::endl;
		}
	}
}

void MainMenu(const IMSSystem& myiMS)
{
	std::cout << std::endl << std::endl;
	std::cout << " ******** MAIN MENU ********" << std::endl;
	int i = 1;
	for (SubMenuList::const_iterator it = menus->begin();
			it != menus->end(); ++it)
	{
		std::cout << " " << i++ << ") " << (*it)->SubMenuDescription() << std::endl;
	}
	std::cout << " q) Quit Program" << std::endl;
}

int main(int argc, char *argv[]) {
	// The system we are going to test
	IMSSystem myiMS;

	std::string s;

	ClearScreen();
	std::cout << WelcomeMsg << std::endl;
	std::cout << "API Version: " << LibVersion::GetVersion()  << std::endl;
	std::cout << "Searching for iMS Systems . . . (please wait)" << std::endl;

	std::unique_ptr<ConnectionList> connList (new ConnectionList());
	std::vector<IMSSystem> fulliMSList = connList->scan();

	if (fulliMSList.size() > 0) {
		std::cout << "Discovered " << fulliMSList.size() << " iMS Systems: " << std::endl;
		int ims_count = 1;
		for (std::vector<IMSSystem>::const_iterator it = fulliMSList.cbegin(); it != fulliMSList.cend(); ++it) {
			std::cout << "  " << ims_count++ << ") " << it->ConnPort() << std::endl;
		}
		for(;;) {
			std::cout << "Select a system to continue ('q' to quit) ... ";
			std::getline( std::cin, s);

			int index;
			if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
				ClearScreen();
				return EXIT_SUCCESS;
			}
			std::istringstream iss(s);
			iss >> index;
			if (!iss.fail()) {
				if ((index > 0) && (index <= (int)(fulliMSList.size()))) {
					std::cout << "Connecting..." << std::endl;
					myiMS = fulliMSList[index - 1];
					myiMS.Connect();
					break;
				} else std::cout << "Try again - ";
			} else std::cout << "Try again - ";
		}
	} else {
		std::cout << "Can't find an iMS System :-(" << std::endl;
		return EXIT_FAILURE;
	}

	// Initialise Menu System
	menus = new SubMenuList(myiMS);

	do
	{
		ClearScreen();
		PrintIMSDetails(myiMS);
		MainMenu(myiMS);
		std::getline( std::cin, s);
		if ((s.find("q") != std::string::npos) || (s.find("Q") != std::string::npos)) {
			break;
		} else {
			int index;
			std::istringstream iss(s);
			iss >> index;
			if (!iss.fail()) {
				if ((index > 0) && (index <= (int)menus->size())) {
					ClearScreen();
					(*menus)[index - 1]->LaunchSubMenu();
				}
			}
		}
	} while(1);

	ClearScreen();
	myiMS.Disconnect();
	return EXIT_SUCCESS;
}
