/*-----------------------------------------------------------------------------
/ Title      : Isomet Tone Buffer Header
/ Project    : Isomet Modular Synthesiser System
/------------------------------------------------------------------------------
/ File       : $URL: http://nutmeg/svn/sw/trunk/09-Isomet/iMS_SDK/API/ToneBuffer/h/ToneBuffer.h $
/ Author     : $Author: dave $
/ Company    : Isomet (UK) Ltd
/ Created    : 2016-02-24
/ Last update: $Date: 2017-09-11 23:55:34 +0100 (Mon, 11 Sep 2017) $
/ Platform   :
/ Standard   : C++11
/ Revision   : $Rev: 300 $
/------------------------------------------------------------------------------
/ Description:
/------------------------------------------------------------------------------
/ Copyright (c) 2016 Isomet (UK) Ltd. All Rights Reserved.
/------------------------------------------------------------------------------
/ Revisions  :
/ Date        Version  Author  Description
/ 2016-02-24  1.0      dc      Created
/
/----------------------------------------------------------------------------*/

///
/// \file ToneBuffer.h
///
/// \brief Class for storing an array of Synthesiser tones
///
/// \author Dave Cowan
/// \date 2016-02-24
/// \since 1.0
/// \ingroup group_ToneBuffer
///

#ifndef IMS_TONEBUFFER_H__
#define IMS_TONEBUFFER_H__

#include "IMSSystem.h"
#include "IEventHandler.h"
#include "IMSTypeDefs.h"
#include "IBulkTransfer.h"
#include "Image.h"
#include "FileSystem.h"

#include <array>

/// \cond LIB_CREATION
#if defined _WIN32 || defined __CYGWIN__
  #ifdef __GNUC__
    #define DLL_EXPORT __attribute__ ((dllexport))
    #define DLL_IMPORT __attribute__ ((dllimport))
  #else
    #define DLL_EXPORT __declspec(dllexport) // Note: actually gcc seems to also supports this syntax.
    #define DLL_IMPORT __declspec(dllimport) // Note: actually gcc seems to also supports this syntax.
  #endif
  #define DLL_LOCAL
#else
  #if __GNUC__ >= 4
    #define DLL_EXPORT __attribute__ ((visibility ("default")))
    #define DLL_IMPORT __attribute__ ((visibility ("default")))
    #define DLL_LOCAL  __attribute__ ((visibility ("hidden")))
  #else
    #define DLL_EXPORT
    #define DLL_IMPORT
    #define DLL_LOCAL
  #endif
#endif

#if defined(_EXPORTING_IMS)
  #define LIBSPEC DLL_EXPORT
  #define LIBLOCAL DLL_LOCAL
  #define EXPIMP_TEMPLATE
#elif defined(_STATIC_IMS)
  #define LIBSPEC
  #define LIBLOCAL
  #define EXPIMP_TEMPLATE
#else
  #define LIBSPEC DLL_IMPORT
  #define LIBLOCAL DLL_LOCAL
  #define EXPIMP_TEMPLATE extern
#endif
/// \endcond

namespace iMS {
	///
	/// \brief TBEntry is synonymous with ImagePoint
	/// An entry in the Tone Buffer contains four FAPs, one per output channel and is therefore
	/// comparable to a single ImagePoint making up one entry in an Image.
	/// \since 1.1
	using TBEntry = ImagePoint;

	///
	/// \class ToneBufferEvents ToneBuffer.h include\ToneBuffer.h
	/// \brief All the different types of events that can be triggered by the ToneBuffer and ToneBufferDownload classes.
	///
	/// Some events contain integer parameter data which can be processed by the IEventHandler::EventAction
	/// derived method
	/// \author Dave Cowan
	/// \date 2016-02-24
	/// \since 1.1
	class LIBSPEC ToneBufferEvents
	{
	public:
		/// \enum Events List of Events raised by the ToneBuffer Class and ToneBuffer Table Downloader
		enum Events {
			/// Event raised when ToneBufferDownload has confirmed that the iMS Controller received all of the ToneBuffer data
			DOWNLOAD_FINISHED,
			/// Event raised each time the ToneBufferDownload class registers an error in the download process
			DOWNLOAD_ERROR,
			Count
		};
	};

	///
    /// \class ToneBuffer ToneBuffer.h include/ToneBuffer.h
    /// \brief An array of 4-channel FAP Tones stored in memory on the Synthesiser
    ///
    /// \author Dave Cowan
    /// \date 2016-02-24
    /// \since 1.1
    ///
	class LIBSPEC ToneBuffer {
	public:
	///
	/// \name Tone Buffer Array
	///
	/// \brief \c TBArray is the internal type definition used for storing a buffer of TBEntry 's in the Image
		using TBArray = std::array < TBEntry, 256>;
	///
	/// \name Iterator Specification
    ///
    /// Use these iterators when you want to work with ranges of Tone Buffer entries stored within a tone buffer.
    /// Iterators can be used to access elements at an arbitrary offset position
    /// relative to the element they point to
    ///
    /// Two types of iterators are supported; both are random access iterators.  Dereferencing const_iterator
    /// yields a reference to a constant element in the ToneBuffer (const TBEntry&).
    ///
    //@{
    /// \brief Iterator defined for user manipulation of internal TBArray
		typedef TBArray::iterator iterator;
    ///
    /// \brief Const Iterator defined for user readback of internal TBArray
		typedef TBArray::const_iterator const_iterator;
    //@}
    ///

    ///
    /// \name Constructors & Destructors
    //@{
		/// \brief Empty Constructor
    ///
		/// \param[in] name The optional descriptive name to apply to the Tone Buffer
		ToneBuffer(const std::string& name = "");
    ///
		/// \brief Fill Constructor
    ///
    /// Use this constructor to generate a Tone Buffer with each entry initialised
    /// to the value of \c tbe
    ///
    /// \param[in] tbe The TBEntry that will fill each of the elements of the TBArray
	/// \param[in] name The optional descriptive name to apply to the Tone Buffer
		/// \since 1.1
		ToneBuffer(const TBEntry& tbe, const std::string& name = "");
    ///
	/// \brief Non-volatile Memory Constructor
	///
	/// Use this constructor to preload the ToneBuffer with data recalled from an entry in the Synthesiser
	/// FileSystem.
	///
	/// \param[in] entry the entry in the FileSystem Table from which to recall a ToneBuffer
	/// \param[in] name The optional descriptive name to apply to the Tone Buffer
		/// \since 1.1
		ToneBuffer(const int entry, const std::string& name = "");

	///
    /// \brief Copy Constructor
    ///
		ToneBuffer(const ToneBuffer &);
    ///
    /// \brief Assignment Constructor
    ///
		ToneBuffer &operator =(const ToneBuffer &);

    ///
    /// \brief Destructor
    ///
		~ToneBuffer();
    ///
    //@}
    ///

    ///
    /// \name ToneBuffer Boundary Iterators
    //@{
    /// \brief Returns an iterator pointing to the first element in the TBArray container.
    ///
    /// \return An iterator to the beginning of the TBArray container.
    /// \since 1.1
		iterator begin();
    /// \brief Returns an iterator referring to the past-the-end element in the TBArray container.
    ///
    /// The past-the-end element is the theoretical element that would follow the last element
    /// in the TBArray container. It does not point to any element, and thus shall not be dereferenced.
    ///
    /// Because the ranges used by functions of the standard library do not include the element
    /// pointed by their closing iterator, this function can be used in combination with
    /// TBArray::begin to specify a range including all the elements in the container.
    ///
    /// \return An iterator to the element past the end of the TBArray.
    /// \since 1.1
		iterator end();
	/// \brief Returns a const_iterator pointing to the first element in the TBArray container.
	/// \return A const_iterator to the beginning of the TBArray container.
	/// \since 1.2.5
		const_iterator begin() const;
	/// \brief Returns a const_iterator referring to the past-the-end element in the TBArray container.
	/// \return A const_iterator to the element past the end of the buffer.
	/// \since 1.2.5
		const_iterator end() const;
	/// \brief Returns a const_iterator pointing to the first element in the TBArray container.
    /// \return A const_iterator to the beginning of the TBArray container.
    /// \since 1.1
		const_iterator cbegin() const;
    /// \brief Returns a const_iterator referring to the past-the-end element in the TBArray container.
    /// \return A const_iterator to the element past the end of the buffer.
    /// \since 1.1
		const_iterator cend() const;
    //@}

    ///
    /// \name TBArray Operators
    //@{
    ///
	/// \brief Random Access to a TBEntry in the TBArray
    ///
    /// The array subscript operator is defined to permit applications to
    /// access a TBEntry at any arbitrary position for readback.
    /// \param[in] idx Integer offset into the TBArray with respect to the first element in the array (ToneBuffer::cbegin())
    /// \return A const reference to a TBEntry.
    /// \since 1.1
		const TBEntry& operator[](std::size_t idx) const;
	/// \brief Random Write Access to a TBEntry in the TBArray
	///
	/// The array subscript operator is defined to permit applications to
	/// access a CompensationPoint at any arbitrary position for modification.
	/// \param[in] idx Integer offset into the TBArray with respect to the first element in the array (ToneBuffer::begin())
	/// \return A reference to a TBEntry.
	/// \since 1.1
		TBEntry& operator[](std::size_t idx);
	///
	/// \brief Equality Operator checks ToneBuffer contents for equivalence
	///
	/// \param[in] rhs A ToneBuffer object to perform the comparison with
	/// \return True if the supplied ToneBuffer is identical to this one.
	/// \since 1.1
		bool operator==(ToneBuffer const& rhs) const;
    //@}

    ///
    /// \name ToneBuffer Size
    //@{
    ///
    /// \brief Returns the number of elements in the ToneBuffer (non-modifiable)
    ///
    /// \return The number of elements in the ToneBuffer
    /// \since 1.1
		const std::size_t Size() const;
	//@}
    ///

	///
	/// \name Tone Buffer Description
	//@{
	///
	/// \brief A string stored with the Tone Buffer to aid human users in identifying the purpose of the buffer
	///
	/// A descriptive string can be set alongside the Tone Buffer to allow users to identify and differentiate
	/// between Tone Buffers without having to browse through the data.  The description is optional, and if,
	/// not used, the description will simply default to null.
	///
		const std::string& Name() const;
		std::string& Name();
	private:
		class Impl;
		Impl * p_Impl;
	};

	///
	/// \class ToneBufferDownload ToneBuffer.h include\ToneBuffer.h
	/// \brief Provides a mechanism for downloading ToneBuffer's to a Synthesiser's LTB memory
	/// \author Dave Cowan
	/// \date 2016-02-24
	/// \since 1.1
	class LIBSPEC ToneBufferDownload : public IBulkTransfer
	{
	public:
		///
		/// \name Constructor & Destructor
		//@{
		///
		/// \brief Constructor for ToneBufferDownload Object
		///
		/// The pre-requisites for an ToneBufferDownload object to be created are:
		///   (1) - an IMSSystem object, representing the configuration of an iMS target to which the ToneBuffer
		/// is to be downloaded.
		///   (2) - a complete ToneBuffer object to download to the iMS target.
		///
		/// ToneBufferDownload stores const references to both.  This means that both must exist before the
		/// ToneBufferDownload object, and both must remain valid (not destroyed) until the ToneBufferDownload
		/// object itself is destroyed.  Because they are stored as references, the IMSSystem and ToneBuffer
		/// objects themselves may be modified after the construction of the ToneBufferDownload object.
		///
		/// Once constructed, the object can neither be copied or assigned to another instance.
		///
		/// \param[in] ims A const reference to the iMS System which is the target for downloading the Image
		/// \param[in] tb A const reference to the ToneBuffer which shall be downloaded to the target
		/// \since 1.1
		ToneBufferDownload(IMSSystem& ims, const ToneBuffer& tb);
		///
		/// \brief Destructor for ToneBufferDownload Object
		~ToneBufferDownload();
		//@}

		/// \name Bulk Transfer Initiation
		//@{
		/// \brief Begins download of entire ToneBuffer to LTB memory on Synthesiser
		/// \since 1.1
		bool StartDownload();
		/// \brief Begins download of partial ToneBuffer to LTB memory on Synthesiser beginning at \c first
		/// TBEntry and continuing until \c last TBEntry (including first but not including last)
		/// \since 1.1
		bool StartDownload(ToneBuffer::const_iterator first, ToneBuffer::const_iterator last);
		/// \brief Downloads a single TBEntry to LTB memory on Synthesiser
		/// \since 1.1
		bool StartDownload(ToneBuffer::const_iterator single);
		/// \brief No Verify is possible. Always returns false
		/// \since 1.1
		bool StartVerify() { return false; };
		/// \brief No Verify is possible. Always return -1
		/// \since 1.1
		int GetVerifyError() { return -1; };
		///
		//@}

		///
		/// \name Event Notifications
		//@{
		///
		/// \brief Subscribe a callback function handler to a given ToneBufferEvents entry
		///
		/// ToneBufferDownload can callback user application code when an event occurs in the
		/// download process.  Supported events are listed under ToneBufferEvents.  The
		/// callback function must inherit from the IEventHandler interface and override
		/// its EventAction() method.
		///
		/// Use this member function call to subscribe a callback function to an ToneBufferEvents entry.
		/// For the period that a callback is subscribed, each time an event in ToneBufferDownload occurs
		/// that would trigger the subscribed ToneBufferEvents entry, the user function callback will be
		/// executed.
		/// \param[in] message Use the ToneBufferEvents::Event enum to specify an event to subscribe to
		/// \param[in] handler A function pointer to the user callback function to execute on the event trigger.
		/// \since 1.1
		void ToneBufferDownloadEventSubscribe(const int message, IEventHandler* handler);
		/// \brief Unsubscribe a callback function handler from a given ToneBufferEvents entry
		///
		/// Removes all links to a user callback function from the Event Trigger map so that any
		/// events that occur in the ToneBufferDownload object following the Unsubscribe request
		/// will no longer execute that function
		/// \param[in] message Use the ToneBufferEvents::Event enum to specify an event to unsubscribe from
		/// \param[in] handler A function pointer to the user callback function that will no longer execute on an event
		/// \since 1.1
		void ToneBufferDownloadEventUnsubscribe(const int message, const IEventHandler* handler);
		//@}

		/// \name Store in Synthesiser Non-Volatile Memory
		//@{
		/// \brief Store ToneBuffer contents to non-volatile memory on the synthesiser
		///
		/// The contents of this ToneBuffer can be stored to an area of non-volatile memory on the
		/// Synthesiser for retrieval at a future time, including after subsequent power cycles.  The
		/// data stored can be used to select between alternative ToneBuffers without needing to
		/// recalculate or download from Software.
		///
		/// The table can be flagged to be used as a default at startup in which case the Synthesiser will
		/// use the contents as a default ToneBuffer program allowing the Synthesiser to be used
		/// with no connection to a host system.
		///
		/// \param[in] def mark the entry as a default and the Synthesiser will attempt 
		/// to program the data to the Local Tone Buffer on power up.
		/// \param[in] FileName a string to tag the download with in the File System Table (limited to 8 chars)
		/// \returns the index in the File System Table where the data was stored or -1 if the operation failed
		/// \since 1.1
		const FileSystemIndex Store(const std::string& FileName, FileDefault def = FileDefault::NON_DEFAULT) const;
		//@}
	private:
		// Makes this object non-copyable
		ToneBufferDownload(const ToneBufferDownload &);
		const ToneBufferDownload &operator =(const ToneBufferDownload &);

		class Impl;
		Impl * p_Impl;
	};

}

#undef EXPIMP_TEMPLATE
#undef LIBSPEC
#endif
