/*-----------------------------------------------------------------------------
/ Title      : iMS Useful Type Defines Header
/ Project    : Isomet Modular Synthesiser System
/------------------------------------------------------------------------------
/ File       : $URL: http://nutmeg/svn/sw/trunk/09-Isomet/iMS_SDK/API/Other/h/IMSTypeDefs.h $
/ Author     : $Author: dave $
/ Company    : Isomet (UK) Ltd
/ Created    : 2015-04-09
/ Last update: $Date: 2017-09-11 23:55:34 +0100 (Mon, 11 Sep 2017) $
/ Platform   :
/ Standard   : C++11
/ Revision   : $Rev: 300 $
/------------------------------------------------------------------------------
/ Description:
/------------------------------------------------------------------------------
/ Copyright (c) 2015 Isomet (UK) Ltd. All Rights Reserved.
/------------------------------------------------------------------------------
/ Revisions  :
/ Date        Version  Author  Description
/ 2015-04-09  1.0      dc      Created
/
/----------------------------------------------------------------------------*/

///
/// \file IMSTypeDefs.h
///
/// \brief Useful Type Definitions for working with iMS Systems
///
///
/// \author Dave Cowan
/// \date 2015-11-03
/// \since 1.0
/// \ingroup group_TypeDefs
///

#ifndef IMS_IMSTYPEDEFS_H__
#define IMS_IMSTYPEDEFS_H__

#include <cmath>
#include <stdexcept>
#include <cstdint>
#include <vector>

/// \cond LIB_CREATION
#if defined _WIN32 || defined __CYGWIN__
  #ifdef __GNUC__
    #define DLL_EXPORT __attribute__ ((dllexport))
    #define DLL_IMPORT __attribute__ ((dllimport))
  #else
    #define DLL_EXPORT __declspec(dllexport) // Note: actually gcc seems to also supports this syntax.
    #define DLL_IMPORT __declspec(dllimport) // Note: actually gcc seems to also supports this syntax.
  #endif
  #define DLL_LOCAL
#else
  #if __GNUC__ >= 4
    #define DLL_EXPORT __attribute__ ((visibility ("default")))
    #define DLL_IMPORT __attribute__ ((visibility ("default")))
    #define DLL_LOCAL  __attribute__ ((visibility ("hidden")))
  #else
    #define DLL_EXPORT
    #define DLL_IMPORT
    #define DLL_LOCAL
  #endif
#endif

#if defined(_EXPORTING_IMS)
  #define LIBSPEC DLL_EXPORT
  #define LIBLOCAL DLL_LOCAL
  #define EXPIMP_TEMPLATE
#elif defined(_STATIC_IMS)
  #define LIBSPEC
  #define LIBLOCAL
  #define EXPIMP_TEMPLATE
#else
  #define LIBSPEC DLL_IMPORT
  #define LIBLOCAL DLL_LOCAL
  #define EXPIMP_TEMPLATE extern
#endif
/// \endcond

namespace iMS
{

	// Forward Declaration
	class IMSSystem;

    ///
    /// \class Frequency IMSTypeDefs.h include/IMSTypeDefs.h
    /// \brief Type Definition for all operations that require a frequency specification
    ///
    /// Internally, the Frequency value is stored as a double precision variable specified in Hertz
    /// \author Dave Cowan
    /// \date 2015-11-03
    /// \since 1.0
    ///
	class LIBSPEC Frequency {
	protected:
		double value;

	public:
    /// \brief Construct a Frequency object from a double argument representing Hertz
    /// \param[in] arg Frequency in Hertz
    /// \since 1.0
		Frequency(double arg = 0.0) : value(arg) {}

    /// \brief Assignment of a double argument in Hertz to an existing Frequency object
    ///
    /// \code
    /// Frequency f;
    /// f = 1000.0;
    /// // f contains 1000Hz
    /// \endcode
    /// \since 1.0
		Frequency& operator = (double arg) {
			value = arg;
			return *this;
		}

    /// \brief Return a double representing the Frequency value in Hertz
    /// \code
    /// kHz f1(1.2);
    /// Frequency f2 = f1();
    /// std::cout << "f2's Frequency is: " << f2() << "Hz" << std::endl;
    /// \endcode
    /// prints:
    /// \code
    /// f2's Frequency is 1200.0Hz
    /// \endcode
    /// \since 1.0
		operator double() const { return value; }

    /// \brief Used internally by the library to convert a Frequency object into an hardware-dependent
    /// integer representation used by the Image for Internal Oscillator frequency
    ///
    /// Not intended for use in application code
    static unsigned int RenderAsPointRate(const IMSSystem&, const Frequency, const bool PrescalerDisable = false);
	};

    ///
    /// \class kHz IMSTypeDefs.h include/IMSTypeDefs.h
    /// \brief Type Definition for all operations that require a frequency specification in kiloHertz
    ///
    /// kHz inherits from Frequency, which internally stores the value in Hertz.
    /// \author Dave Cowan
    /// \date 2015-11-03
    /// \since 1.0
    ///
	class LIBSPEC kHz : public Frequency {
	public:
    /// \brief Construct a kHz object from a double argument representing kiloHertz
    /// \param[in] arg Frequency in kiloHertz
    /// \since 1.0
		kHz(double arg) : Frequency(arg * 1000.0) {};

    /// \brief Assignment of a double argument in kiloHertz to an existing Frequency object
    ///
    /// \code
    /// kHz f;
    /// f = 1.0;
    /// // f contains 1000Hz
    /// \endcode
    /// \since 1.3
		kHz& operator = (double arg) {
			value = 1000.0 * arg;
			return *this;
		}

    /// \brief Return a double representing the Frequency value in kiloHertz
    /// \code
    /// Frequency f1(3750.0);
    /// kHz f2 = f1();
    /// std::cout << "f2's Frequency is: " << f2() << "kHz" << std::endl;
    /// \endcode
    /// prints:
    /// \code
    /// f2's Frequency is 3.75kHz
    /// \endcode
    /// \since 1.0
		operator double() const { return (value / 1000.0); }
	};

    ///
    /// \class MHz IMSTypeDefs.h include/IMSTypeDefs.h
    /// \brief Type Definition for all operations that require a frequency specification in MegaHertz
    ///
    /// MHz inherits from Frequency, which internally stores the value in Hertz.
    /// \author Dave Cowan
    /// \date 2015-11-03
    /// \since 1.0
    ///
	class LIBSPEC MHz : public Frequency {
	public:
    /// \brief Construct a MHz object from a double argument representing MegaHertz
    /// \param[in] arg Frequency in MegaHertz
    /// \since 1.0
		MHz(double arg) : Frequency(arg * 1000000.0) {};

    /// \brief Assignment of a double argument in MegaHertz to an existing Frequency object
    ///
    /// \code
    /// MHz f;
    /// f = 1.0;
    /// // f contains 1,000,000Hz
    /// \endcode
    /// \since 1.3
		MHz& operator = (double arg) {
			value = 1000000.0 * arg;
			return *this;
		}

    /// \brief Return a double representing the Frequency value in MegaHertz
    /// \code
    /// Frequency f1(1234567.0);
    /// MHz f2 = f1();
    /// std::cout << "f2's Frequency is: " << f2() << "MHz" << std::endl;
    /// \endcode
    /// prints:
    /// \code
    /// f2's Frequency is 1.234567MHz
    /// \endcode
    /// \since 1.0
		operator double() const { return (value / 1000000.0); }

    /// \brief Used internally by the library to convert a Frequency object into a hardware-dependent
    /// integer representation used by the Image for RF Output frequency
    ///
    /// Not intended for use in application code
		static unsigned int RenderAsImagePoint(const IMSSystem&, const MHz);
	};

    ///
    /// \class Percent IMSTypeDefs.h include/IMSTypeDefs.h
    /// \brief Type Definition for all operations that require a percentage specification
    ///
    /// Internally, the Percent value is stored as a double precision variable and is
    /// bounds-limited to 0.0 <= Percent <= 100.0.
    /// \author Dave Cowan
    /// \date 2015-11-03
    /// \since 1.0
    ///
	class LIBSPEC Percent{
		double value;

	public:
	/// \brief Default Constructor assigns 0.0%
	/// \since 1.1
		Percent() : value(0.0) {};
    /// \brief Construct a Percent object from a double argument and check its value is within
    /// the range 0.0 <= arg <= 100.0.  If not, the object is still constructed, but the value is
    /// clipped to the upper or lower bound.
    /// \param[in] arg The percentage value
    /// \since 1.0
		Percent(double arg) : value(arg) {
			if (arg < 0.0) value = 0.0;
			if (arg > 100.0) value = 100.0;
		}

    /// \brief Assignment of a double argument in percent to an existing Percent object.
    ///
    /// The double argument of the assigner must be within the range 0.0 <= arg <= 100.0 else it
    /// will be limited to those bounds.
    /// \code
    /// // In a group of 7 children, 3 of them have dark hair
    /// Percent ChildrenWithDarkHair = (3.0 / 7.0) * 100.0;
    /// std::cout << ChildrenWithDarkHair << "% of the group have dark hair" << std::endl;
    /// \endcode
    /// prints:
    /// \code
    /// 42.8571% of the group have dark hair
    /// \endcode
    /// \since 1.0
		Percent& operator = (double arg) {
			if (arg < 0.0) value = 0.0;
			else if (arg > 100.0) value = 100.0;
			else value = arg;
			return *this;
		}

    /// \brief Return a double representing the Percent object's value
    /// \since 1.0
		operator double() const { return value; }

    /// \brief Used internally by the library to convert a Percent object into a hardware-dependent
    /// integer representation used by the Image for RF Output amplitude
    ///
    /// Not intended for use in application code
		static unsigned int RenderAsImagePoint(const IMSSystem&, const Percent);

    /// \brief Used internally by the library to convert a Percent object into a hardware-dependent
    /// integer representation used by the Compensation Table for Compensation amplitude
    ///
    /// Not intended for use in application code
		static unsigned int RenderAsCompensationPoint(const IMSSystem&, const Percent);

	/// \brief Used internally by the library to convert a Percent object into a hardware-dependent
	/// integer representation used by the Calibration Tone for Single Tone amplitude
	///
	/// Not intended for use in application code
	///
	/// \since 1.1.0
		static unsigned int RenderAsCalibrationTone(const IMSSystem&, const Percent);
	};

    ///
    /// \class Degrees IMSTypeDefs.h include/IMSTypeDefs.h
    /// \brief Type Definition for all operations that require an angle specification in degrees
    ///
    /// Internally, the Degrees value is stored as a double precision variable and is
    /// limited to sit within the range 0.0 <= Percent < 360.0.
    /// \author Dave Cowan
    /// \date 2015-11-03
    /// \since 1.0
    ///
	class LIBSPEC Degrees {
		double value;

	public:
    /// \brief Construct a Degrees object from a double argument and check its value is within
    /// the range 0.0 <= arg < 360.0.  If not, the object is still constructed, but the value is
    /// wrapped around to fit within the range.
    /// \param[in] arg The percentage value
    /// \since 1.0
		Degrees(double arg) {
			value = std::fmod(arg, 360.0);
		}

    /// \brief Assignment of a double argument in degrees to an existing Degrees object.
    ///
    /// The double argument of the assigner must be within the range 0.0 <= arg < 360.0 else it
    /// will be wrapped around to fit within the range.
    /// \code
    /// // needed for PI
    /// #define _USE_MATH_DEFINES
    /// #include <iostream>
    /// #include <cmath>
    ///
    /// Degrees phase = atan2(1.0, -1.0) * (360.0 / 2 * M_PI);
    /// std::cout << "The arctangent for [x=-1, y=1] is " << phase << " degrees" << std::endl;
    /// \endcode
    /// prints:
    /// \code
    /// The arctangent for [x=-1, y=1] is 135.000000 degrees
    /// \endcode
    /// \since 1.0
		Degrees& operator = (double arg) {
			value = std::fmod(arg, 360.0);
			return *this;
		}

    /// \brief Return a double representing the Degrees object's value
    /// \since 1.0
		operator double() const { return value; }

    /// \brief Used internally by the library to convert a Degrees object into a hardware-dependent
    /// integer representation used by the Image for RF Output phase
    ///
    /// Not intended for use in application code
		static unsigned int RenderAsImagePoint(const IMSSystem&, const Degrees);

    /// \brief Used internally by the library to convert a Degrees object into a hardware-dependent
    /// integer representation used by the Compensation Table for channel phase increment
    ///
    /// Not intended for use in application code
		static unsigned int RenderAsCompensationPoint(const IMSSystem&, const Degrees);

	/// \brief Used internally by the library to convert a Degrees object into a hardware-dependent
	/// integer representation used by the Calibration Tone for channel phase increment
	///
	/// Not intended for use in application code
	///
	/// \since 1.1.0
		static unsigned int RenderAsCalibrationTone(const IMSSystem&, const Degrees);
	};

    /// \struct FAP IMSTypeDefs.h include/IMSTypeDefs.h
    /// \brief FAP (Frequency/Amplitude/Phase) triad stores the instantaneous definition of a single RF output
    ///
    /// The FAP struct, also known as a triad, stores one frequency (in MHz), one amplitude (Percent) and one
    /// phase (Degrees) value which uniquely specifies the instantaneous output of any one RF channel output.
    ///
    /// 4 FAP's make up a single ImagePoint, one per RF channel, and sequences of ImagePoints then
    /// make up an Image.
    /// \author Dave Cowan
    /// \date 2015-11-03
    /// \since 1.0
    ///
	struct LIBSPEC FAP
	{
    /// The RF Channel Output Frequency
		MHz freq;
    /// The RF Channel Output Amplitude
		Percent ampl;
    /// The RF Channel Output Phase
		Degrees phase;

	/// \brief Default construct a FAP object with zero data
		FAP() : freq(0.0), ampl(0.0), phase(0.0) {};
    /// \brief Construct a FAP object from raw double precision input data
		FAP(double f, double a, double p) : freq(f), ampl(a), phase(p) {};
    /// \brief Construct a FAP object from pre-existing MHz, Percent and Degrees objects
		FAP(MHz f, Percent a, Degrees p) : freq(f), ampl(a), phase(p) {};

	/// \name Equality Operators
	//@{
	/// \brief Equality operators compare FAPs against each other
	/// \since 1.1.0
	///
		bool operator==(const FAP &other) const;
		bool operator!=(const FAP &other) const;
	};
	//@}

    /// \class RFChannel IMSTypeDefs.h include/IMSTypeDefs.h
    /// \brief Type that represents the integer values 1, 2, 3 and 4, one each for the RF Channels of an iMS Synthesiser
    ///
    /// The type is used to ensure that incorrect channel specifications cannot be passed to functions
    /// requiring an argument referencing an RF output channel.  Attempting to use an integer outside
    /// the range 1 <= arg <= 4 will result in RFChannel = 1 and an invalid_argument exception being thrown.
    /// \throws std::invalid_argument("Invalid RF Channel Number") Attempted to use an integer specification
    /// not tied to an RF Output Channel
    /// \author Dave Cowan
    /// \date 2015-11-03
    /// \since 1.0
    ///
	class LIBSPEC RFChannel {
		int value;

	public:
	/// \brief Default construct an RF Channel object initialised to the first RF Channel
	/// \since 1.1
		RFChannel() : value(1) {};
    /// \brief Construct an RF Channel object and check that it is being created with an integer value
    /// within the range 1 <= arg <= 4.  If not, the object is still constructed, but the RF Channel
    /// value is set to 1 and an invalid_argument exception is thrown
    /// \param[in] arg The channel specification
    /// \throws std::invalid_argument("Invalid RF Channel Number") Attempted to use an integer specification
    /// not tied to an RF Output Channel
    /// \since 1.0
		RFChannel(int arg) : value(arg) {
			if (arg < 1 || arg > 4) {
				value = 1;
				throw std::invalid_argument("Invalid RF Channel Number");
			}
		}

    /// \brief Assignment of an integer argument to an existing RF Channel object
    ///
    /// Checks that it is being created with an integer value
    /// within the range 1 <= arg <= 4.  If not, the object is still constructed, but the RF Channel
    /// value is set to 1 and an invalid_argument exception is thrown
    /// \param[in] arg The channel specification
    /// \throws std::invalid_argument("Invalid RF Channel Number") Attempted to use an integer specification
    /// not tied to an RF Output Channel
    /// \since 1.0
		RFChannel& operator = (int arg) {
			if (arg < 1 || arg > 4) {
				value = 1;
				throw std::invalid_argument("Invalid RF Channel Number");
			}
			else value = arg;
			return *this;
		}

	/// \name
	//@{
	/// \brief Prefix and Postfix operators for (dec)incrementing through channels
	/// \since 1.1
		RFChannel& operator++() {
			if (value<4) value++;
			return *this;
		}

		RFChannel operator++(int) {
			RFChannel temp = *this;
			++(*this);
			return temp;
		}

		RFChannel& operator--() {
			if (value>1) value--;
			return *this;
		}

		RFChannel operator--(int) {
			RFChannel temp = *this;
			--(*this);
			return temp;
		}
	//@}

	/// \brief Return an integer representing the RF Channel that the object references
    /// \since 1.0
		operator int() const { return value; }
	};

}

#undef EXPIMP_TEMPLATE
#undef LIBSPEC
#endif
