/*-----------------------------------------------------------------------------
/ Title      : Isomet Image Project Header
/ Project    : Isomet Modular Synthesiser System
/------------------------------------------------------------------------------
/ File       : $URL: http://nutmeg/svn/sw/trunk/09-Isomet/iMS_SDK/API/ImageOps/h/ImageProject.h $
/ Author     : $Author: dave $
/ Company    : Isomet (UK) Ltd
/ Created    : 2015-04-09
/ Last update: $Date: 2017-09-11 23:55:34 +0100 (Mon, 11 Sep 2017) $
/ Platform   :
/ Standard   : C++11
/ Revision   : $Rev: 300 $
/------------------------------------------------------------------------------
/ Description: The ImageProject class organises Image's, ImageGroup's,
/ CompensationFunction's and ToneBuffer's in a single entity.  The application
/ software may use any or all of the containers according to need.  An efficient
/ custom file format has been defined, both in compressed (.iip or IsometImageProject)
/ and uncompressed (.xml) formats and ImageProject's may be saved to and loaded
/ from disk in these formats.  The class will also read in ImageProject's
/ generated by Isomet's previous generation Image software, the iHHS ImageFileGenerator.
/------------------------------------------------------------------------------
/ Copyright (c) 2015 Isomet (UK) Ltd. All Rights Reserved.
/------------------------------------------------------------------------------
/ Revisions  :
/ Date        Version  Author  Description
/ 2015-04-09  1.0      dc      Created
/
/----------------------------------------------------------------------------*/

///
/// \file ImageProject.h
///
/// \brief Classes for organising Images and associated data
///
/// \author Dave Cowan
/// \date 2015-11-03
/// \since 1.0
/// \ingroup group_Image
///

#ifndef IMS_IMAGEPROJECT_H__
#define IMS_IMAGEPROJECT_H__

#include "IMSTypeDefs.h"
#include "Containers.h"
#include "Image.h"
#include "Compensation.h"
#include "ToneBuffer.h"

/// \cond LIB_CREATION
#if defined _WIN32 || defined __CYGWIN__
  #ifdef __GNUC__
    #define DLL_EXPORT __attribute__ ((dllexport))
    #define DLL_IMPORT __attribute__ ((dllimport))
  #else
    #define DLL_EXPORT __declspec(dllexport) // Note: actually gcc seems to also supports this syntax.
    #define DLL_IMPORT __declspec(dllimport) // Note: actually gcc seems to also supports this syntax.
  #endif
  #define DLL_LOCAL
#else
  #if __GNUC__ >= 4
    #define DLL_EXPORT __attribute__ ((visibility ("default")))
    #define DLL_IMPORT __attribute__ ((visibility ("default")))
    #define DLL_LOCAL  __attribute__ ((visibility ("hidden")))
  #else
    #define DLL_EXPORT
    #define DLL_IMPORT
    #define DLL_LOCAL
  #endif
#endif

#if defined(_EXPORTING_IMS)
  #define LIBSPEC DLL_EXPORT
  #define LIBLOCAL DLL_LOCAL
  #define EXPIMP_TEMPLATE
#elif defined(_STATIC_IMS)
  #define LIBSPEC
  #define LIBLOCAL
  #define EXPIMP_TEMPLATE
#else
  #define LIBSPEC DLL_IMPORT
  #define LIBLOCAL DLL_LOCAL
  #define EXPIMP_TEMPLATE extern
#endif
/// \endcond

namespace iMS {

	///
	/// \class ImageGroupList Image.h include/Image.h
	/// \brief A List of ImageGroup's used as a container by ImageProject 
	///
	/// \date 2016-11-09
	/// \since 1.3
	///
	class LIBSPEC ImageGroupList : public ListBase < ImageGroup >
	{};

	///
	/// \class CompensationFunctionList Image.h include/Image.h
	/// \brief A List of CompensationFunction's used as a container by ImageProject
	///
	/// \date 2016-11-09
	/// \since 1.3
	///
	class LIBSPEC CompensationFunctionList : public ListBase < CompensationFunction >
	{};

	///
	/// \class ToneBufferList Image.h include/Image.h
	/// \brief A List of ToneBuffer's used as a container by ImageProject
	///
	/// \date 2016-11-09
	/// \since 1.3
	///
	class LIBSPEC ToneBufferList : public ListBase < ToneBuffer >
	{};

	///
	/// \class ImageProject Image.h include/Image.h
	/// \brief An ImageProject allows the user to organise their data and store it on the host computer.
	///
	/// An ImageProject permits the user to organise and contain all of their data relating to a specific use case for the iMS.
	/// Any number of ImageGroup's, CompensationFunction's, ToneBuffer's and/or Free Images (individual Images not associated in
	/// an ImageGroup and with no ImageSequence) may be help in an ImageProject.  The data is stored on disk in an efficient
	/// custom file format that may be unzipped by the user and inspected in an XML file viewer, if wished.
	///
	/// A simple use case for the ImageProject is to contain a single Image, kept in the FreeImageContainer, and to load/save that
	/// Image to disk.  A more complex use case might be to hold different ImageGroup's for different purposes that are used as the
	/// raw data for a processing task that requires some other software to select between Image's and ImageSequence's according to
	/// some system parameter.
	///
	/// \date 2016-11-09
	/// \since 1.3
	///
	class LIBSPEC ImageProject
	{
	public:
		/// \name Constructors & Destructor
		//@{
		/// \brief Default Constructor
		///
		/// Creates an empty ImageProject.
		ImageProject();
		/// \brief Implicit Load From File Constructor
		///
		/// Calls the default constructor, then ImageProject::Load() to initialise the object from the filesystem
		/// param[in] fileName String pointing to a valid .iip or .xml file on the host filesystem to load into the ImageProject
		ImageProject(const std::string& fileName);
		///
		//@}


		/// \name Image Group Container
		//@{
		/// \brief Set Accessor for the Image Group Container
		///
		/// Use this function to modify, add and remove ImageGroup's from the ImageProject.
		ImageGroupList& ImageGroupContainer();
		/// \brief Get Accessor for the Image Group Container
		const ImageGroupList& ImageGroupContainer() const;
		///
		//@}

		/// \name Compensation Function Container
		//@{
		/// \brief Set Accessor for the Compensation Function Container
		///
		/// Use this function to modify, add and remove CompensationFunction's from the ImageProject.
		CompensationFunctionList& CompensationFunctionContainer();
		/// \brief Get Accessor for the Compensation Function Container
		const CompensationFunctionList& CompensationFunctionContainer() const;
		///
		//@}

		/// \name Tone Buffer Container
		//@{
		/// \brief Set Accessor for the Tone Buffer Container
		///
		/// Use this function to modify, add and remove ToneBuffer's from the ImageProject.
		ToneBufferList& ToneBufferContainer();
		/// \brief Get Accessor for the Tone Buffer Container
		const ToneBufferList& ToneBufferContainer() const;
		///
		//@}

		/// \name Free Image Container
		//@{
		/// \brief Set Accessor for the Free Image Container
		///
		/// Use this function to modify, add and remove individual Image's from the ImageProject.
		/// The Free Image Container is configured as an ImageGroup to allow user software to 
		/// access its Image's using the same function calls as when working with an ImageGroup
		/// object.  In this regard, the FreeImageContainer may be regarded as a "superset" of the
		/// ImageGroup that permits save/load to disk.
		ImageGroup& FreeImageContainer();
		/// \brief Get Accessor for the Tone Buffer Container
		const ImageGroup& FreeImageContainer() const;
		///
		//@}

		/// \name FileSystem Functions
		//@{
		/// \brief Save ImageProject to host disk
		///
		/// Stores all data in the ImageProject containers to a custom file format on disk.  The
		/// preferred file extension is ".iip" for Isomet Image Project.  If the passed file name ends 
		/// in .xml however, the function will save the data to an uncompressed XML type file which 
		/// can be read in any XML file viewer.
		///
		/// \warning XML files saved this way can be very large!
		///
		/// If the passed file name ends in neither .xml nor .iip, the function will save the data in
		/// compressed format as if it ended in .iip but respecting the user's choice of file name.
		/// \param[in] fileName String representing the full path of a file to save data to, ending in .xml (uncompressed) or .iip (compressed)
		bool Save(const std::string& fileName);
		/// \brief Load ImageProject from host disk
		///
		/// First clears the ImageProject of any existing content.  Then it will try to read in data from the
		/// provided file, work out whether it is compressed or uncompressed, and populate the ImageProject
		/// containers from the file data.  The function is fully backwards compatible with all previous
		/// versions of the SDK and should therefore be capable of reading in any file ever generated by the SDK.
		/// It is also compatible with ImageProject files generated by the previous generation of Isomet Image
		/// software, the iHHS ImageFile Generator
		/// \param[in] fileName String representing the full path of a file to load fata from.
		bool Load(const std::string& fileName);
		///
		//@}

		/// \brief Reset ImageProject to empty state
		///
		/// Removes all entries from all of the containers
		void Clear();
	private:
		class Impl;
		Impl *p_Impl;
	};
}

#undef EXPIMP_TEMPLATE
#undef LIBSPEC
#endif
